<?php

namespace App\Libraries;

use Request;
use Response;
class Rest
{
    private const AUTH_CREDENTIALS = [];
    private const AUTH_DIGEST_REALM = 'Restricted area';
    private const AUTH_DIGEST_QOP = 'auth';
    private const UNAUTHORIZED = ["code" => 401, "title" => "HTTP/1.1 401 Unauthorized Request", "message" => "The client request has not been completed because it lacks valid authentication credentials for the requested resource."];
    private const NOT_FOUND = ["code" => 404, "title" => "HTTP/1.1 404 Not Found", "message" => "The server cannot find the requested resource."];
    private const METHOD_NOT_ALLOWED = ["code" => 405, "title" => "HTTP/1.1 405 Method Not Allowed", "message" => "The server knows the request method, but the target resource doesn't support this method."];
    private const NOT_ACCEPTABLE = ["code" => 406, "title" => "HTTP/1.1 406 Not Acceptable", "message" => "The server cannot produce a response matching the list of acceptable values defined in the request's proactive content negotiation headers, and that the server is unwilling to supply a default representation."];
    private const VALIDATION = ["code" => 422, "title" => "HTTP/1.1 422 Unprocessable Entity", "message" => "The server understands the content type of the request entity, and the syntax of the request entity is correct, but it was unable to process the contained instructions."];
    private const LOGIN_TIMEOUT = ["code" => 440, "title" => "HTTP/1.1 440 Login Timeout", "message" => "You have either logged out or you were automatically logged out for security purposes."];
    private const SUCCESS = ["code" => 200, "title" => "HTTP/1.1 200 OK", "message" => "OK"];

    private string $error = "";
    private array $requiredFields = [];
    private array $optionalFields = [];

    function authorizedRequest()
    {
        if (in_array($this->platform, ["web", "onsite"])) {
            if (isset($_SERVER['HTTP_X_API_KEY']) === false)
                $this->unauthorizedHeader();
            if ($_SERVER['HTTP_X_API_KEY'] !== REST_API_AUTH_CREDENTIALS[$this->platform]["key"])
                $this->unauthorizedHeader();
            return true;
        }
        $digest = "";
        if (isset($_SERVER['PHP_AUTH_DIGEST']))
            $digest = $_SERVER['PHP_AUTH_DIGEST'];
        elseif (isset($_SERVER['HTTP_AUTHORIZATION']))
            if (strpos(strtolower($_SERVER['HTTP_AUTHORIZATION']), 'digest') === 0)
                $digest = substr($_SERVER['HTTP_AUTHORIZATION'], 7);

        $digest = $this->http_digest_parse($digest);

        if (empty($digest))
            $this->unauthorizedHeader();
        else {
            $nonce = $digest['nonce'];
            $platform = self::AUTH_CREDENTIALS[$this->platform];
            $A1 = md5($platform['un'] . ":" . self::AUTH_DIGEST_REALM . ":" . $platform['pw']);
            $A2 = md5($_SERVER['REQUEST_METHOD'] . ":" . $digest['uri']);
            $validResponse = md5($A1 . ':' . $nonce . ':' . $digest['nc'] . ':' . $digest['cnonce'] . ':' . $digest['qop'] . ':' . $A2);

            if ($digest['response'] !== $validResponse)
                $this->unauthorized();
        }
    }

    private function unauthorizedHeader()
    {
        $nonce = dcEncode(date("YmdHis") . uniqid());
        header('HTTP/1.1 ' . self::UNAUTHORIZED["code"] . ' Unauthorized');
        header('WWW-Authenticate: Digest realm="' . self::AUTH_DIGEST_REALM . '",qop="' . self::AUTH_DIGEST_QOP . '",nonce="' . $nonce . '",opaque="' . md5(self::AUTH_DIGEST_QOP) . '"');

    }

    private function http_digest_parse($digest)
    {
        $needed_parts = array('nonce' => 1, 'nc' => 1, 'cnonce' => 1, 'qop' => 1, 'username' => 1, 'uri' => 1, 'response' => 1);
        $data = array();
        $keys = implode('|', array_keys($needed_parts));

        preg_match_all('@(' . $keys . ')=(?:([\'"])([^\2]+?)\2|([^\s,]+))@', $digest, $matches, PREG_SET_ORDER);

        foreach ($matches as $m) {
            $data[$m[1]] = $m[3] ? $m[3] : $m[4];
            unset($needed_parts[$m[1]]);
        }
        return $needed_parts ? false : $data;
    }

    function unauthorized($error = "")
    {
        $response = ["head" => ["status" => false, "code" => self::UNAUTHORIZED["code"], "message" => $error ?? self::UNAUTHORIZED["message"]]];
        $this->response($response, self::UNAUTHORIZED["code"]);

    }

    private function response($response, $headerCde)
    {
        return response()->json($response)->header('X-Powered-By', "Dreamcast");
    }

    function json($data = [])
    {
        $response = ["head" => ["status" => true, "code" => self::SUCCESS["code"]]];
        $response["body"] = $data;
        return $this->response($response, self::SUCCESS["code"]);
    }

    function setDefaultFailureMessage($error): Rest
    {
        $this->error = $error;
        return $this;
    }

    function methodNotAllowed($error = ""): \Illuminate\Http\JsonResponse
    {
        $response = ["head" => ["status" => false, "code" => self::METHOD_NOT_ALLOWED["code"], "message" => $error ?? self::METHOD_NOT_ALLOWED["message"]]];
        return $this->response($response, self::METHOD_NOT_ALLOWED["code"]);

    }

    function setRequiredFields(...$fields): Rest
    {
        $this->requiredFields = $fields;
        return $this;
    }

    function setOptionalFields(...$fields): Rest
    {
        $this->optionalFields = $fields;
        return $this;
    }

    function validateRequest()
    {
        $requestStatus = true;
        $requestFields = array_keys($this->request->data);

        foreach ($this->requiredFields as $unused => $value) {
            if (($key = array_search($value, $requestFields)) !== false) {
                unset($requestFields[$key]);
                continue;
            }
            $requestStatus = false;
            break;
        }
        if ($requestStatus === true) {
            foreach ($this->optionalFields as $unused => $value) {
                if (($key = array_search($value, $requestFields)) !== false) {
                    unset($requestFields[$key]);
                }
            }
            $requestStatus = count($requestFields) === 0;
        }

        if ($requestStatus === false)
            $this->returnRequestFailure();

    }

    function invalidRequest($error = "")
    {
        $response = ["head" => ["status" => false, "code" => self::NOT_ACCEPTABLE["code"], "message" => $error ?? self::NOT_ACCEPTABLE["message"]]];
        return $this->response($response, self::NOT_ACCEPTABLE["code"]);
    }

    function notFound()
    {
        $response = ["head" => ["status" => false, "code" => self::NOT_FOUND["code"], "message" => self::NOT_FOUND["message"]]];
        return $this->response($response, self::NOT_FOUND["code"]);
    }

    function loginTimeout()
    {
        $response = ["head" => ["status" => false, "code" => self::LOGIN_TIMEOUT["code"], "message" => self::LOGIN_TIMEOUT["message"]]];
        return $this->response($response, self::LOGIN_TIMEOUT["code"]);
    }

    function validatorErrors($validator, $customValidationErrors = [])
    {
        $errorBag = $validator->errors()->toArray();
        $errors = [];
        foreach ($errorBag as $field => $error) {
//            $errors[] = ["field" => $field, "error" => $error[0]];
            $errors[$field] = $error[0];
        }
        foreach ($customValidationErrors as $field => $error) {
//            $errors[] = ["field" => $field, "error" => $error];
            $errors[$field] = $error;
        }
        $response = ["head" => ["status" => false, "code" => self::VALIDATION["code"], "message" => self::VALIDATION["message"]]];
        $response["body"] = $errors;
        return $this->response($response, self::VALIDATION["code"]);
    }

    function fail($message)
    {
        $body = ["message" => $message, "status" => false];
        $response = ["head" => $body, "body" => $body];
        $response["head"]["code"] = self::SUCCESS["code"];
        return $this->response($response, self::SUCCESS["code"]);
    }

    function success($message)
    {
        $body = ["message" => $message, "status" => true];
        $response = ["head" => $body, "body" => $body];
        $response["head"]["code"] = self::SUCCESS["code"];
        return $this->response($response, self::SUCCESS["code"]);
    }

    protected function parseRawHttpRequest(array &$body)
    {
        if (isset($_SERVER['CONTENT_TYPE'])) {
            $input = file_get_contents('php://input');

            preg_match('/boundary=(.*)$/', $_SERVER['CONTENT_TYPE'], $matches);

            if (!count($matches)) {
                parse_str(urldecode($input), $body);
                return $body;
            }

            $boundary = $matches[1];

            $blocks = preg_split("/-+$boundary/", $input);
            array_pop($blocks);

            foreach ($blocks as $unused => $block) {
                if (empty($block))
                    continue;
                list($mime, $content) = explode("\r\n\r\n", $block, 2);
                if (preg_match('/name="([^"]*)"/i', $mime, $match) && isset($match[1])) {
                    $body[$match[1]] = str_replace(array("\r", "\n"), "", $content);
                } else {
                    $body[] = $content;
                }
            }
        }
    }

    private function is($method)
    {
        $serverElementList = [
            'HTTP_X_REQUESTED_WITH' => 'XMLHttpRequest',
            'HTTP_USER_AGENT' => ['ios', 'android'],
            'REQUEST_METHOD' => ["GET", "POST", "PUT", "PATCH", "DELETE"],
            'HTTP_DC_TIMEZONE' => [-720, 840],
//            'HTTP_DC_OS' => ["android", "ios"],
//            'HTTP_DC_DEVICE' => ["mobile","tablet","desktop","tv"],
        ];

        foreach ($serverElementList as $key => $value) {
            if ($this->platform === "onsite" && $key === "HTTP_USER_AGENT")
                continue;

            if (!isset($_SERVER[$key])) {
                $this->returnFailureMethodNotAllowed();
                break;
            }

            if ($key == "HTTP_X_REQUESTED_WITH" && strtolower($_SERVER[$key]) !== strtolower($value)) {
                $this->returnFailureMethodNotAllowed();
                break;
            }
            if ($key == "HTTP_DC_TIMEZONE" && $_SERVER[$key] < $value[0] && $_SERVER[$key] > $value[1]) {
                $this->returnFailureMethodNotAllowed();
                break;
            }
            if ($key == "REQUEST_METHOD" && ($_SERVER[$key] !== $method || !in_array($method, $value))) {
                $this->returnFailureMethodNotAllowed();
                break;
            }
            if ($this->platform !== "web" && $key == "HTTP_USER_AGENT") {
                $httpUserAgent = strtolower($_SERVER[$key]);
                if (!in_array($httpUserAgent, $value) && stripos($httpUserAgent, "postmanruntime") === false)
                    $this->returnFailureMethodNotAllowed();
                break;
            }

        }
    }
}
